from typing import TypedDict, Required, Literal


class SchemaDict(TypedDict, total=False):
    pass


# ---------------------------------------------------------------------------------------
# Types


class FileSystemCapability(SchemaDict, total=False):
    readTextFile: bool
    writeTextFile: bool


# https://agentclientprotocol.com/protocol/schema#clientcapabilities
class ClientCapabilities(SchemaDict, total=False):
    fs: FileSystemCapability
    terminal: bool


# https://agentclientprotocol.com/protocol/schema#implementation
class Implementation(SchemaDict, total=False):
    name: Required[str]
    title: str | None
    version: Required[str]


# https://agentclientprotocol.com/protocol/schema#promptcapabilities
class PromptCapabilities(SchemaDict, total=False):
    audio: bool
    embeddedContent: bool
    image: bool


# https://agentclientprotocol.com/protocol/schema#agentcapabilities
class AgentCapabilities(SchemaDict, total=False):
    loadSession: bool
    promptCapabilities: PromptCapabilities


class AuthMethod(SchemaDict, total=False):
    description: str | None
    id: Required[str]
    name: Required[str]


# https://agentclientprotocol.com/protocol/schema#envvariable
class EnvVariable(SchemaDict, total=False):
    _meta: dict
    name: Required[str]
    value: Required[str]


# https://agentclientprotocol.com/protocol/schema#terminalexitstatus
class TerminalExitStatus(SchemaDict, total=False):
    _meta: dict
    exitCode: int | None
    signal: str | None


# https://agentclientprotocol.com/protocol/schema#mcpserver
class McpServer(SchemaDict, total=False):
    args: list[str]
    command: str
    env: list[EnvVariable]
    name: str


# https://modelcontextprotocol.io/specification/2025-06-18/server/resources#annotations
class Annotations(SchemaDict, total=False):
    audience: list[str]
    priority: float
    lastModified: str


class TextContent(SchemaDict, total=False):
    type: Required[str]
    text: Required[str]
    annotations: Annotations


class ImageContent(SchemaDict, total=False):
    type: Required[str]
    data: Required[str]
    mimeType: Required[str]
    url: str
    annotations: Annotations


class AudioContent(SchemaDict, total=False):
    type: Required[str]
    data: Required[str]
    mimeType: Required[str]
    Annotations: Annotations


class EmbeddedResourceText(SchemaDict, total=False):
    uri: Required[str]
    text: Required[str]
    mimeType: str


class EmbeddedResourceBlob(SchemaDict, total=False):
    uri: Required[str]
    blob: Required[str]
    mimeType: str


# https://agentclientprotocol.com/protocol/content#embedded-resource
class EmbeddedResourceContent(SchemaDict, total=False):
    type: Required[str]
    resource: EmbeddedResourceText | EmbeddedResourceBlob


class ResourceLinkContent(SchemaDict, total=False):
    annotations: Annotations | None
    description: str | None
    mimeType: str | None
    name: Required[str]
    size: int | None
    title: str | None
    type: Required[str]
    uri: Required[str]


# https://agentclientprotocol.com/protocol/schema#contentblock
type ContentBlock = (
    TextContent
    | ImageContent
    | AudioContent
    | EmbeddedResourceContent
    | ResourceLinkContent
)


# https://agentclientprotocol.com/protocol/schema#param-user-message-chunk
class UserMessageChunk(SchemaDict, total=False):
    content: Required[ContentBlock]
    sessionUpdate: Required[Literal["user_message_chunk"]]


class AgentMessageChunk(SchemaDict, total=False):
    content: Required[ContentBlock]
    sessionUpdate: Required[Literal["agent_message_chunk"]]


class AgentThoughtChunk(SchemaDict, total=False):
    content: Required[ContentBlock]
    sessionUpdate: Required[Literal["agent_thought_chunk"]]


class ToolCallContentContent(SchemaDict, total=False):
    content: Required[ContentBlock]
    type: Required[Literal["content"]]


# https://agentclientprotocol.com/protocol/schema#param-diff
class ToolCallContentDiff(SchemaDict, total=False):
    newText: Required[str]
    oldText: str | None
    path: Required[str]
    type: Required[Literal["diff"]]


class ToolCallContentTerminal(SchemaDict, total=False):
    terminalId: Required[str]
    type: Required[Literal["terminal"]]


# https://agentclientprotocol.com/protocol/schema#toolcallcontent
type ToolCallContent = (
    ToolCallContentContent | ToolCallContentDiff | ToolCallContentTerminal
)

# https://agentclientprotocol.com/protocol/schema#toolkind
type ToolKind = Literal[
    "read",
    "edit",
    "delete",
    "move",
    "search",
    "execute",
    "think",
    "fetch",
    "switch_mode",
    "other",
]

type ToolCallStatus = Literal["pending", "in_progress", "completed", "failed"]


class ToolCallLocation(SchemaDict, total=False):
    line: int | None
    path: Required[str]


type ToolCallId = str


# https://agentclientprotocol.com/protocol/schema#toolcall
class ToolCall(SchemaDict, total=False):
    _meta: dict
    content: list[ToolCallContent]
    kind: ToolKind
    locations: list[ToolCallLocation]
    rawInput: dict
    rawOutput: dict
    sessionUpdate: Required[Literal["tool_call"]]
    status: ToolCallStatus
    title: Required[str]
    toolCallId: Required[ToolCallId]


# https://agentclientprotocol.com/protocol/schema#toolcallupdate
class ToolCallUpdate(SchemaDict, total=False):
    _meta: dict
    content: list[ToolCallContent] | None
    kind: ToolKind | None
    locations: list | None
    rawInput: dict
    rawOutput: dict
    sessionUpdate: Required[Literal["tool_call_update"]]
    status: ToolCallStatus | None
    title: str | None
    toolCallId: Required[ToolCallId]


# https://agentclientprotocol.com/protocol/schema#param-tool-call
# Use in the session/request_permission call (not the same as ToolCallUpdate)
class ToolCallUpdatePermissionRequest(SchemaDict, total=False):
    _meta: dict
    content: list[ToolCallContent] | None
    kind: ToolKind | None
    locations: list | None
    rawInput: dict
    rawOutput: dict
    status: ToolCallStatus | None
    title: str | None
    toolCallId: Required[ToolCallId]


class PlanEntry(SchemaDict, total=False):
    content: Required[str]
    priority: Literal["high", "medium", "low"]
    status: Literal["pending", "in_progress", "completed"]


type SessionModeId = str


# https://agentclientprotocol.com/protocol/schema#sessionmode
class SessionMode(SchemaDict, total=False):
    _meta: dict
    description: str | None
    id: Required[SessionModeId]
    name: Required[str]


class SessionModeState(SchemaDict, total=False):
    _meta: dict
    availableModes: Required[list[SessionMode]]
    currentModeId: Required[SessionModeId]


type ModelId = str


# https://agentclientprotocol.com/protocol/schema#modelinfo
class ModelInfo(SchemaDict, total=False):
    _meta: dict
    description: str | None
    modelId: Required[ModelId]
    name: Required[str]


# https://agentclientprotocol.com/protocol/schema#sessionmodelstate
class SessionModelState(SchemaDict, total=False):
    _meta: dict
    availableModels: Required[list[ModelInfo]]
    currentModelId: Required[ModelId]


# https://agentclientprotocol.com/protocol/schema#param-plan
class Plan(SchemaDict, total=False):
    entries: Required[list[PlanEntry]]
    sessionUpdate: Required[Literal["plan"]]


class AvailableCommandInput(SchemaDict, total=False):
    hint: Required[str]


class AvailableCommand(SchemaDict, total=False):
    description: Required[str]
    input: AvailableCommandInput | None
    name: Required[str]


class AvailableCommandsUpdate(SchemaDict, total=False):
    availableCommands: Required[list[AvailableCommand]]
    sessionUpdate: Required[Literal["available_commands_update"]]


class CurrentModeUpdate(SchemaDict, total=False):
    currentModeId: Required[str]
    sessionUpdate: Required[Literal["current_mode_update"]]


type SessionUpdate = (
    UserMessageChunk
    | AgentMessageChunk
    | AgentThoughtChunk
    | ToolCall
    | ToolCallUpdate
    | Plan
    | AvailableCommandsUpdate
    | CurrentModeUpdate
)


class SessionNotification(TypedDict, total=False):
    sessionId: str
    update: SessionUpdate


type PermissionOptionKind = Literal[
    "allow_once", "allow_always", "reject_once", "reject_always"
]
type PermissionOptionId = str


class PermissionOption(TypedDict, total=False):
    _meta: dict
    kind: Required[PermissionOptionKind]
    name: Required[str]
    optionId: Required[PermissionOptionId]


class OutcomeCancelled(TypedDict, total=False):
    outcome: Literal["cancelled"]


class OutcomeSelected(TypedDict, total=False):
    optionId: Required[PermissionOptionId]
    outcome: Literal["selected"]


# https://agentclientprotocol.com/protocol/schema#requestpermissionoutcome
type RequestPermissionOutcome = OutcomeSelected | OutcomeCancelled

# ---------------------------------------------------------------------------------------
# RPC responses


class InitializeResponse(SchemaDict, total=False):
    agentCapabilities: AgentCapabilities
    authMethods: list[AuthMethod]
    protocolVersion: Required[int]


# https://agentclientprotocol.com/protocol/schema#newsessionresponse
class NewSessionResponse(SchemaDict, total=False):
    _meta: object
    sessionId: Required[str]
    # Unstable from here
    models: SessionModelState | None
    modes: SessionModeState | None


class SessionPromptResponse(SchemaDict, total=False):
    stopReason: Required[
        Literal[
            "end_turn",
            "max_tokens",
            "max_turn_requests",
            "refusal",
            "cancelled",
        ]
    ]


# https://agentclientprotocol.com/protocol/schema#requestpermissionresponse
class RequestPermissionResponse(TypedDict, total=False):
    _meta: dict
    outcome: Required[RequestPermissionOutcome]


# https://agentclientprotocol.com/protocol/schema#createterminalresponse
class CreateTerminalResponse(TypedDict, total=False):
    _meta: dict
    terminalId: Required[str]


# https://agentclientprotocol.com/protocol/schema#killterminalcommandresponse
class KillTerminalCommandResponse(TypedDict, total=False):
    _meta: dict


# https://agentclientprotocol.com/protocol/schema#terminaloutputresponse
class TerminalOutputResponse(TypedDict, total=False):
    _meta: dict
    exitStatus: TerminalExitStatus | None
    output: Required[str]
    truncated: Required[bool]


# https://agentclientprotocol.com/protocol/schema#releaseterminalresponse
class ReleaseTerminalResponse(TypedDict, total=False):
    _meta: dict


# https://agentclientprotocol.com/protocol/schema#waitforterminalexitresponse
class WaitForTerminalExitResponse(TypedDict, total=False):
    _meta: dict
    exitCode: int | None
    signal: str | None


# https://agentclientprotocol.com/protocol/schema#setsessionmoderesponse
class SetSessionModeResponse(TypedDict, total=False):
    meta: dict


# ---------------------------------------------------------------------------------------



import base64
from pathlib import Path

from toad.acp import protocol
from toad.prompt.extract import extract_paths_from_prompt
from toad.prompt.resource import load_resource, ResourceError


def build(project_path: Path, prompt: str) -> list[protocol.ContentBlock]:
    """Build the prompt structure and extract paths with the @ syntax.

    Args:
        project_path: The project root.
        prompt: The prompt text.

    Returns:
        A list of content blocks.
    """
    prompt_content: list[protocol.ContentBlock] = []

    prompt_content.append({"type": "text", "text": prompt})
    for path, _, _ in extract_paths_from_prompt(prompt):
        if path.endswith("/"):
            continue
        try:
            resource = load_resource(project_path, Path(path))
        except ResourceError:
            # TODO: How should this be handled?
            continue
        uri = f"file://{resource.path.absolute().resolve()}"
        if resource.text is not None:
            prompt_content.append(
                {
                    "type": "resource",
                    "resource": {
                        "uri": uri,
                        "text": resource.text,
                        "mimeType": resource.mime_type,
                    },
                }
            )
        elif resource.data is not None:
            prompt_content.append(
                {
                    "type": "resource",
                    "resource": {
                        "uri": uri,
                        "blob": base64.b64encode(resource.data).decode("utf-8"),
                        "mimeType": resource.mime_type,
                    },
                }
            )

    return prompt_content



from __future__ import annotations

from math import ceil
from typing import ClassVar
from time import monotonic

from textual.color import Color
from textual.reactive import var
from textual.content import Content
from textual.style import Style
from textual.timer import Timer
from textual.widgets import Static


class FutureText(Static):
    """Text which appears one letter at time, like the movies."""

    DEFAULT_CSS = """
    FutureText {
        width: auto;
        height: 1;
        text-wrap: nowrap;
        text-align: center;
        color: $primary;
        &>.future-text--cursor {
            color: $primary;
        }
    }
    """
    ALLOW_SELECT = False
    COMPONENT_CLASSES = {"future-text--cursor"}

    BARS: ClassVar[list[str]] = ["▉", "▊", "▋", "▌", "▍", "▎", "▏", " "]
    text_offset = var(0)

    def __init__(
        self,
        text_list: list[Content],
        *,
        speed: float = 16.0,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.text_list = text_list
        self.speed = speed
        self.start_time = monotonic()
        super().__init__(name=name, id=id, classes=classes)
        self._update_timer: Timer | None = None

    @property
    def text(self) -> Content:
        return self.text_list[self.text_offset % len(self.text_list)]

    @property
    def time(self) -> float:
        return monotonic() - self.start_time

    def on_mount(self) -> None:
        self.start_time = monotonic()
        self.set_interval(1 / 60, self._update_text)

    def _update_text(self) -> None:
        if not self.is_attached or not self.screen.is_active:
            return
        text = self.text + " "
        speed_time = self.time * self.speed
        progress, fractional_progress = divmod(speed_time, 1)
        end = progress >= len(text)
        cursor_progress = 0 if end else int(fractional_progress * 8)
        text = text[: ceil(progress)]

        bar_character = self.BARS[7 - cursor_progress]

        cursor_styles = self.get_component_styles("future-text--cursor")
        cursor_style = Style(foreground=cursor_styles.color)
        reverse_cursor_style = cursor_style + Style(reverse=True)

        # Fade in last character
        fade_style = Style(
            foreground=Color.blend(
                cursor_styles.background, cursor_styles.color, ceil(fractional_progress)
            )
        )

        fade_text = Content.assemble(
            text[:-1],
            ((text[-1].plain if text else " "), fade_style),
        )

        if speed_time >= 1:
            text = Content.assemble(
                fade_text,
                (bar_character, reverse_cursor_style),
                (bar_character, cursor_style),
                " " * (len(self.text) + 1 - len(fade_text)),
            )
        self.update(text, layout=False)

        if progress > len(text) + 10 * 5:
            self.text_offset += 1
            self.start_time = monotonic()


if __name__ == "__main__":
    from textual.app import App, ComposeResult

    TEXT = [Content("Thinking..."), Content("Working hard..."), Content("Nearly there")]

    class TextApp(App):
        CSS = """
        Screen {
            padding: 2 4;
            FutureText {
                width: auto;
                max-width: 1fr;
                height: auto;
               
            }
        }
        """

        def compose(self) -> ComposeResult:
            yield FutureText(TEXT)

    TextApp().run()



from dataclasses import dataclass

from textual.app import ComposeResult
from textual.content import Content
from textual.layout import Layout
from textual.reactive import reactive
from textual import containers
from textual.widgets import Static

from toad.pill import pill
from toad.widgets.strike_text import StrikeText


class NonSelectableStatic(Static):
    ALLOW_SELECT = False


class Plan(containers.Grid):
    BORDER_TITLE = "Plan"
    DEFAULT_CLASSES = "block"
    DEFAULT_CSS = """
    Plan {
        background: black 20%;
        height: auto;
        padding: 0 1;
        margin: 0 1 1 1;
        border: tall transparent;
        
        grid-size: 2;
        grid-columns: auto 1fr;
        grid-rows: auto;
        height: auto;        
    
        .plan {
            color: $text-secondary;
        }
        .status {
            padding: 0 0 0 0;
            color: $text-secondary;
        }
        .priority {
            padding: 0 0 0 0;
        }
        .status.status-completed {
            color: $text-success;
            text-style: bold;
        }
        .status-pending {
            opacity: 0.8;
        }
    }

    """

    @dataclass(frozen=True)
    class Entry:
        """Information about an entry in the Plan."""

        content: Content
        priority: str
        status: str

    entries: reactive[list[Entry] | None] = reactive(None, recompose=True)

    LEFT = Content.styled("▌", "$error-muted on transparent r")

    PRIORITIES = {
        "high": pill("H", "$error-muted", "$text-error"),
        "medium": pill("M", "$warning-muted", "$text-warning"),
        "low": pill("L", "$primary-muted", "$text-primary"),
    }

    def __init__(
        self,
        entries: list[Entry],
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.newly_completed: set[Plan.Entry] = set()
        super().__init__(name=name, id=id, classes=classes)
        self.set_reactive(Plan.entries, entries)

    def watch_entries(self, old_entries: list[Entry], new_entries: list[Entry]) -> None:
        entry_map = {entry.content: entry for entry in old_entries}
        newly_completed: set[Plan.Entry] = set()
        for entry in new_entries:
            old_entry = entry_map.get(entry.content, None)
            if (
                old_entry is not None
                and entry.status == "completed"
                and entry.status != old_entry.status
            ):
                newly_completed.add(entry)
        self.newly_completed = newly_completed

    def compose(self) -> ComposeResult:
        if not self.entries:
            yield Static("No plan yet", classes="-no-plan")
            return
        for entry in self.entries:
            classes = f"priority-{entry.priority} status-{entry.status}"
            yield NonSelectableStatic(
                self.render_status(entry.status),
                classes=f"status {classes}",
            )

            yield (
                strike_text := StrikeText(
                    entry.content,
                    classes=f"plan {classes}",
                )
            )
            if entry in self.newly_completed or (
                not self.is_mounted and entry.status == "completed"
            ):
                self.call_after_refresh(strike_text.strike)

    def render_status(self, status: str) -> Content:
        if status == "completed":
            return Content.from_markup("✔ ")
        elif status == "pending":
            return Content.styled("⏲ ")
        elif status == "in_progress":
            return Content.from_markup("⮕")
        return Content()


if __name__ == "__main__":
    from textual.app import App

    entries = [
        Plan.Entry(
            Content.from_markup(
                "Build the best damn UI for agentic coding in the terminal"
            ),
            "high",
            "in_progress",
        ),
        Plan.Entry(Content.from_markup("???"), "medium", "in_progress"),
        Plan.Entry(
            Content.from_markup("[b]Profit[/b]. Retire to Costa Rica"),
            "low",
            "pending",
        ),
    ]

    new_entries = [
        Plan.Entry(
            Content.from_markup(
                "Build the best damn UI for agentic coding in the terminal"
            ),
            "high",
            "completed",
        ),
        Plan.Entry(Content.from_markup("???"), "medium", "in_progress"),
        Plan.Entry(
            Content.from_markup("[b]Profit[/b]. Retire to Costa Rica"),
            "low",
            "pending",
        ),
    ]

    class PlanApp(App):
        BINDINGS = [("space", "strike")]

        CSS = """
        Screen {
            align: center middle;
        }
        """

        def compose(self) -> ComposeResult:
            yield Plan(entries)

        def action_strike(self) -> None:
            self.query_one(Plan).entries = new_entries

    app = PlanApp()
    app.run()



from __future__ import annotations


import asyncio
from functools import lru_cache
from operator import itemgetter
import os
from pathlib import Path
import re2 as re
from typing import Sequence


from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual import work
from textual import getters
from textual import containers
from textual import events
from textual.actions import SkipAction

from textual.reactive import var, Initialize
from textual.content import Content, Span
from textual.strip import Strip
from textual.widget import Widget
from textual import widgets
from textual.widgets import OptionList, Input, DirectoryTree
from textual.widgets.option_list import Option


from toad import directory
from toad.fuzzy import FuzzySearch
from toad.messages import Dismiss, InsertPath, PromptSuggestion
from toad.path_filter import PathFilter
from toad.widgets.project_directory_tree import ProjectDirectoryTree


class PathFuzzySearch(FuzzySearch):
    @classmethod
    @lru_cache(maxsize=1024)
    def get_first_letters(cls, candidate: str) -> frozenset[int]:
        return frozenset(
            {
                0,
                *[match.start() + 1 for match in re.finditer(r"/", candidate)],
            }
        )

    def score(self, candidate: str, positions: Sequence[int]) -> float:
        """Score a search.

        Args:
            search: Search object.

        Returns:
            Score.
        """
        first_letters = self.get_first_letters(candidate)
        # This is a heuristic, and can be tweaked for better results
        # Boost first letter matches
        offset_count = len(positions)
        score: float = offset_count + len(first_letters.intersection(positions))

        # if 0 in first_letters:
        #     score += 1

        groups = 1
        last_offset, *offsets = positions
        for offset in offsets:
            if offset != last_offset + 1:
                groups += 1
            last_offset = offset

        # Boost to favor less groups
        normalized_groups = (offset_count - (groups - 1)) / offset_count
        score *= 1 + (normalized_groups * normalized_groups)
        return score


class FuzzyInput(Input):
    """Adds a Content placeholder to fuzzy input.

    TODO: Add this ability to Textual.
    """

    HELP = """\
## Fuzzy search

Type a few characters from the file you are searching for.

The search is *fuzzy*, and will match characters that aren't neccesarily next to each other—only the order matters.
"""

    def render_line(self, y: int) -> Strip:
        if y == 0 and not self.value:
            placeholder = Content.from_markup(self.placeholder).expand_tabs()
            placeholder = placeholder.stylize(self.visual_style)
            placeholder = placeholder.stylize(
                self.get_visual_style("input--placeholder")
            )
            if self.has_focus:
                cursor_style = self.get_visual_style("input--cursor")
                if self._cursor_visible:
                    # If the placeholder is empty, there's no characters to stylise
                    # to make the cursor flash, so use a single space character
                    if len(placeholder) == 0:
                        placeholder = Content(" ")
                    placeholder = placeholder.stylize(cursor_style, 0, 1)

            strip = Strip(placeholder.render_segments())
            return strip

        return super().render_line(y)


class PathSearch(containers.VerticalGroup):

    BINDING_GROUP_TITLE = "Path search"

    CURSOR_BINDING_GROUP = Binding.Group(description="Move selection")
    BINDINGS = [
        Binding(
            "up", "cursor_up", "Cursor up", group=CURSOR_BINDING_GROUP, priority=True
        ),
        Binding(
            "down",
            "cursor_down",
            "Cursor down",
            group=CURSOR_BINDING_GROUP,
            priority=True,
        ),
        Binding("enter", "submit", "Insert path", priority=True, show=False),
        Binding("escape", "dismiss", "Dismiss", priority=True, show=False),
        Binding("tab", "switch_picker", "Switch picker", priority=True, show=False),
    ]

    def get_fuzzy_search(self) -> FuzzySearch:
        return PathFuzzySearch(case_sensitive=False)

    root: var[Path] = var(Path("./"))
    paths: var[list[Path]] = var(list)
    highlighted_paths: var[list[Content]] = var(list)
    filtered_path_indices: var[list[int]] = var(list)
    loaded = var(False)
    filter = var("")
    fuzzy_search: var[FuzzySearch] = var(Initialize(get_fuzzy_search))
    show_tree_picker: var[bool] = var(False)

    option_list = getters.query_one(OptionList)
    tree_view = getters.query_one(ProjectDirectoryTree)
    input = getters.query_one(Input)

    def __init__(self, root: Path) -> None:
        super().__init__()
        self.root = root

    def compose(self) -> ComposeResult:
        with widgets.ContentSwitcher(initial="path-search-fuzzy"):
            with containers.VerticalGroup(id="path-search-fuzzy"):
                yield FuzzyInput(
                    compact=True, placeholder="fuzzy search \t[r]▌tab▐[/r] tree view"
                )
                yield OptionList()
            with containers.VerticalGroup(id="path-search-tree"):
                yield widgets.Static(
                    Content.from_markup(
                        "tree view \t[r]▌tab▐[/] fuzzy search"
                    ).expand_tabs(),
                    classes="message",
                )
                yield ProjectDirectoryTree(self.root).data_bind(path=PathSearch.root)

    def on_mount(self) -> None:
        tree = self.tree_view
        tree.guide_depth = 2
        tree.center_scroll = True

    def watch_show_tree_picker(self, show_tree_picker: bool) -> None:
        content_switcher = self.query_one(widgets.ContentSwitcher)
        content_switcher.current = (
            "path-search-tree" if show_tree_picker else "path-search-fuzzy"
        )
        if show_tree_picker:
            self.tree_view.focus()

        else:
            self.input.focus()

    def action_switch_picker(self) -> None:
        self.show_tree_picker = not self.show_tree_picker

    async def search(self, search: str) -> None:
        if not search:
            self.option_list.set_options(
                [
                    Option(highlighted_path, highlighted_path.plain)
                    for highlighted_path in self.highlighted_paths[:100]
                ],
            )
            return

        fuzzy_search = self.fuzzy_search
        fuzzy_search.cache.grow(len(self.paths))
        scores: list[tuple[float, Sequence[int], Content]] = [
            (
                *fuzzy_search.match(search, highlighted_path.plain),
                highlighted_path,
            )
            for highlighted_path in self.highlighted_paths
        ]

        scores = sorted(
            [score for score in scores if score[0]], key=itemgetter(0), reverse=True
        )
        scores = scores[:20]

        def highlight_offsets(path: Content, offsets: Sequence[int]) -> Content:
            return path.add_spans(
                [Span(offset, offset + 1, "underline") for offset in offsets]
            )

        self.option_list.set_options(
            [
                Option(
                    highlight_offsets(path, offsets) if index < 20 else path,
                    id=path.plain,
                )
                for index, (score, offsets, path) in enumerate(scores)
            ]
        )
        with self.option_list.prevent(OptionList.OptionHighlighted):
            self.option_list.highlighted = 0
        self.post_message(PromptSuggestion(""))

    def action_cursor_down(self) -> None:
        if self.show_tree_picker:
            self.tree_view.action_cursor_down()
        else:
            self.option_list.action_cursor_down()

    def action_cursor_up(self) -> None:
        if self.show_tree_picker:
            self.tree_view.action_cursor_up()
        else:
            self.option_list.action_cursor_up()

    def action_dismiss(self) -> None:
        self.post_message(Dismiss(self))

    def on_show(self) -> None:
        self.focus()

    def focus(self, scroll_visible: bool = False) -> Self:
        if self.show_tree_picker:
            return self.tree_view.focus(scroll_visible=scroll_visible)
        else:
            return self.input.focus(scroll_visible=scroll_visible)

    def on_descendant_blur(self, event: events.DescendantBlur) -> None:
        if self.show_tree_picker:
            if event.widget == self.tree_view:
                self.post_message(Dismiss(self))
        else:
            if event.widget == self.input:
                self.post_message(Dismiss(self))

    @on(DirectoryTree.NodeHighlighted)
    def on_node_highlighted(self, event: DirectoryTree.NodeHighlighted) -> None:
        event.stop()

        dir_entry = event.node.data
        if dir_entry is not None:
            try:
                path = Path(dir_entry.path).resolve().relative_to(self.root.resolve())
            except ValueError:
                # Being defensive here, shouldn't occur
                return
            tree_path = str(path)
            self.post_message(PromptSuggestion(tree_path))

    @on(DirectoryTree.FileSelected)
    def on_file_selected(self, event: DirectoryTree.FileSelected) -> None:
        event.stop()

        dir_entry = event.node.data
        if dir_entry is not None:
            try:
                path = Path(dir_entry.path).resolve().relative_to(self.root.resolve())
            except ValueError:
                return
            tree_path = str(path)
            self.post_message(InsertPath(tree_path))
            self.post_message(Dismiss(self))

    @on(Input.Changed)
    async def on_input_changed(self, event: Input.Changed):
        await self.search(event.value)

    @on(OptionList.OptionHighlighted)
    async def on_option_list_changed(self, event: OptionList.OptionHighlighted):
        event.stop()
        if event.option:
            self.post_message(PromptSuggestion(event.option.id))

    @on(OptionList.OptionSelected)
    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        self.action_submit()

    def action_submit(self):
        if self.show_tree_picker:
            raise SkipAction()

        elif (highlighted := self.option_list.highlighted) is not None:
            option = self.option_list.options[highlighted]
            if option.id:
                self.post_message(InsertPath(option.id))
                self.post_message(Dismiss(self))

    def get_path_filter(self, project_path: Path) -> PathFilter:
        """Get a PathFilter insance for the give project path.

        Args:
            project_path: Project path.

        Returns:
            `PathFilter` object.
        """
        path_filter = PathFilter.from_git_root(project_path)
        return path_filter

    def reset(self) -> None:
        """Reset and focus input."""
        self.input.clear()
        self.input.focus()

    @work(exclusive=True)
    async def refresh_paths(self):
        self.loading = True
        root = self.root

        try:
            path_filter = await asyncio.to_thread(self.get_path_filter, root)
            self.tree_view.path_filter = path_filter
            self.tree_view.clear()
            await self.tree_view.reload()
            paths = await directory.scan(
                root, path_filter=path_filter, add_directories=True
            )

            paths = [path.absolute() for path in paths]
            self.root = root
            self.paths = paths
        finally:
            self.loading = False

    def get_loading_widget(self) -> Widget:
        from textual.widgets import LoadingIndicator

        return LoadingIndicator()

    def highlight_path(self, path: str) -> Content:
        content = Content.styled(path, "dim $text")
        if os.path.split(path)[-1].startswith("."):
            return content
        content = content.highlight_regex("[^/]*?$", style="not dim $text-primary")
        content = content.highlight_regex(r"\.[^/]*$", style="italic")
        return content

    def watch_paths(self, paths: list[Path]) -> None:
        self.option_list.highlighted = None

        def path_display(path: Path) -> str:
            try:
                is_directory = path.is_dir()
            except OSError:
                is_directory = False
            if is_directory:
                return str(path.relative_to(self.root)) + "/"
            else:
                return str(path.relative_to(self.root))

        display_paths = sorted(map(path_display, paths), key=str.lower)
        self.highlighted_paths = [self.highlight_path(path) for path in display_paths]
        self.option_list.set_options(
            [
                Option(highlighted_path, id=highlighted_path.plain)
                for highlighted_path in self.highlighted_paths
            ][:100]
        )
        with self.option_list.prevent(OptionList.OptionHighlighted):
            self.option_list.highlighted = 0
        self.post_message(PromptSuggestion(""))



from textual import containers
from textual.app import ComposeResult
from textual import widgets


from toad.acp.protocol import ToolCallContent


class ACPToolCallContent(containers.VerticalGroup):

    def __init__(
        self,
        content: list[ToolCallContent],
        *,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._content = content
        super().__init__(id=id, classes=classes)

    def compose(self) -> ComposeResult:
        for content in self._content:
            match content:
                case {
                    "type": "content",
                    "content": {
                        "type": "text",
                        "text": text,
                    },
                }:
                    yield widgets.Markdown(text)
                case {
                    "type": "diff",
                    "oldText": old_text,
                    "newText": new_text,
                    "path": path,
                }:
                    from toad.widgets.diff_view import DiffView

                    yield DiffView(path, path, old_text or "", new_text)



from dataclasses import dataclass

from typing import Literal

from textual.content import Content
from textual.widget import Widget
from textual.message import Message


class WorkStarted(Message):
    """Work has started."""


class WorkFinished(Message):
    """Work has finished."""


@dataclass
class HistoryMove(Message):
    """Getting a new item form history."""

    direction: Literal[-1, +1]
    shell: bool
    body: str


@dataclass
class UserInputSubmitted(Message):
    body: str
    shell: bool = False
    auto_complete: bool = False


@dataclass
class PromptSuggestion(Message):
    suggestion: str


@dataclass
class Dismiss(Message):
    widget: Widget

    @property
    def control(self) -> Widget:
        return self.widget


@dataclass
class InsertPath(Message):
    path: str


@dataclass
class ChangeMode(Message):
    mode_id: str | None


@dataclass
class Flash(Message):
    """Request a message flash.

    Args:
        Message: Content of flash.
        style: Semantic style.
        duration: Duration in seconds or `None` for default.
    """

    content: str | Content
    style: Literal["default", "warning", "success", "error"]
    duration: float | None = None


class ProjectDirectoryUpdated(Message):
    """The project directory may may changed."""



from typing import Iterable
from textual.widgets import Markdown

from toad.menus import MenuItem


class MarkdownNote(Markdown):
    def get_block_menu(self) -> Iterable[MenuItem]:
        return
        yield

    def get_block_content(self, destination: str) -> str | None:
        return self.source
